@@ -25,6 +25,7 @@ from paste.request import construct_url
 
 __all__ = [
     'default_map_environ',
+    'generator_tracker',
     'make_app',
     ]
 
@@ -35,7 +36,7 @@ default_error_template='''<html>
 <h1>Oops!</h1>
 <p>Something broke while generating the page.
 Please try again in a few minutes, and if the problem persists file
-a bug or contact customer support. PLease quote OOPS-ID
+a bug or contact customer support. Please quote OOPS-ID
 <strong>%(id)s</strong>
 </p></body></html>'''
 
@@ -48,7 +49,7 @@ default_map_environ = {
 
 def make_app(app, config, template=default_error_template,
         content_type='text/html', error_render=None, oops_on_status=None,
-        map_environ=None):
+        map_environ=None, tracker=None):
     """Construct a middleware around app that will forward errors via config.
 
     Any errors encountered by the app will be forwarded to config and an error
@@ -81,6 +82,11 @@ def make_app(app, config, template=default_error_template,
         present map into the OOPS context when generating an OOPS. The value of
         the key determines the name given in the OOPS context. If None is passed
         the default_map_environ is used. Pass {} in to entirely disable mapping.
+    :param tracker: A factory function to create a tracker. Trackers are used
+        to allow variations on the WSGI environment to still use oops_wsgi. 
+        See generator_tracker for the reference tracker used in regular WSGI
+        environments. generator_tracker is used by default or when
+        tracker=None.
     :return: A WSGI app.
     """
     def oops_middleware(environ, start_response):
@@ -143,35 +149,77 @@ def make_app(app, config, template=default_error_template,
                 state['response'] = (status, headers)
             return oops_write
         try:
-            iterator = iter(app(environ, oops_start_response))
-            for bytes in iterator:
+            def ensure_start_response():
                 if 'write' not in state:
                     status, headers = state.pop('response')
                     # Signal that we have called start_response
                     state['write'] = start_response(status, headers)
-                yield bytes
+            def on_exception(exc_info):
+                report = config.create(make_context(exc_info=exc_info))
+                ids = config.publish(report)
+                if not ids or 'write' in state:
+                    # No OOPS generated, no oops publisher, or we have already
+                    # transmitted the wrapped apps headers - either way we can't
+                    # replace the content with a clean error, so let the wsgi
+                    # server figure it out.
+                    raise
+                headers = [('Content-Type', content_type)]
+                headers.append(('X-Oops-Id', str(report['id'])))
+                start_response(
+                    '500 Internal Server Error', headers, exc_info=exc_info)
+                del exc_info
+                if error_render is not None:
+                    return error_render(report)
+                else:
+                    return template % report
+            if tracker is None:
+                tracker_factory = generator_tracker
+            else:
+                tracker_factory = tracker
+            return tracker_factory(
+                ensure_start_response, ensure_start_response, on_exception,
+                app(environ, oops_start_response))
         except socket.error:
             raise
-        except GeneratorExit:
-            raise
         except Exception:
             exc_info = sys.exc_info()
-            report = config.create(make_context(exc_info=exc_info))
-            ids = config.publish(report)
-            if not ids or 'write' in state:
-                # No OOPS generated, no oops publisher, or we have already
-                # transmitted the wrapped apps headers - either way we can't
-                # replace the content with a clean error, so let the wsgi
-                # server figure it out.
-                raise
-            headers = [('Content-Type', content_type)]
-            headers.append(('X-Oops-Id', str(report['id'])))
-            start_response(
-                '500 Internal Server Error', headers, exc_info=exc_info)
-            del exc_info
-            if error_render is not None:
-                yield error_render(report)
-            else:
-                yield template % report
+            return [on_exception(exc_info)]
 
     return oops_middleware
+
+
+def generator_tracker(on_first_bytes, on_finish, on_error, app_body):
+    """A wrapper for generators that calls the OOPS hooks as needed.
+
+    :param on_first_bytes: Called as on_first_bytes() when the first bytes from
+        the app body are available but before they are yielded.
+    :param on_finish: Called as on_finish() when the app body is fully
+        consumed.
+    :param on_error: Called as on_error(sys.exc_info()) if a handleable error
+        has occured while consuming the generator. Errors like GeneratorExit
+        are not handleable.
+    :param app_body: The iterable body for the WSGI app. This may be a simple
+        list or a generator - it is merely known to meet the iterator protocol.
+    """
+    try:
+        called_first = False
+        for bytes in app_body:
+            if not called_first:
+                called_first = True
+                on_first_bytes()
+            yield bytes
+        on_finish()
+    except socket.error:
+        # start_response, which iteration can trigger a call into, may raise
+        # socket.error when writing if the client has disconnected: thats not
+        # an OOPS condition. This does potentially mask socket.error issues in
+        # the appserver code, so we may want to change this to callback to
+        # determine if start_response has been called upstream, and if so, to
+        # still generate an OOPS.
+        raise
+    except GeneratorExit:
+        # Python 2.4
+        raise
+    except Exception:
+        exc_info = sys.exc_info()
+        yield on_error(exc_info)
